(* Time-stamp: <modified the 18/06/2020 (at 15:34) by Erwan Jahier> *)
(** {1 The Algorithm programming Interface.} *)
(** 
    {1 What's need to be provided by users.}
    A SASA process is an instance of an algorithm defined via this
    module.   

    In order to define a SASA algorithm, one need to provide an enable
    and a step function:
   - the enable function states which actions can be triggered;
   - the step function triggers an enable action.

 The enable and the step functions should be of the  type
 [ 's enable_fun ] and [ 's step_fun ] respectively.
 *)

type 's neighbor
type action  = string 
type 's enable_fun = 's -> 's neighbor list -> action list
type 's step_fun   = 's -> 's neighbor list -> action -> 's
(** The first argument  holds the current state of the  process.  As it is
    polymorphic (['s]),  algorithms designers  can put anything  they need
    into this state (an integer,  a structure, etc.).  The only constraint
    is that all algorithms should use the same type.
    
    The  second  argument  holds  the  process  neighbors. Note that  SASA
    processes,  that  live  in  graph  nodes, can  only  access  to  their
    immediate  neighbors.  From  each neighbor,  a process  can access  to
    various information thanks to the functions below.
*)

type 's state_init_fun = int -> string -> 's
(**  The  initial  value of  the  local  state  can  be set  using  an
   initialization function that takes as input 2 arguments: the number
   of  node neighbors,  and  the pid.   Anonymous  algorithms are  not
   supposed to use the pid.  *)

(** Returns the processes local state *) 
val state : 's neighbor -> 's

(** Returns the  channel number, that let this neighbor  access to the
   content of the current process, if its neighbor can access it.  The
   channel number is  the rank,starting at 0, in  the neighbors' list.
   Returns -1 if the neighbor can not access to the process, which may
   happen in directed graphs.

   An  algorithm  that  uses  reply,   and  not  the  pid,  is  called
   semi-anonymous. It is called anonymous if it can access none. *)
val reply : 's neighbor -> int

(**  Returns the  weight of  the  edge from  the current  node to  the
   neighbor.  Note  that "weight"  is an edge  (dot) attribute.   1 is
   returned if not weight is set in the graph. *)
val weight : 's neighbor -> int

val print_neighbor: 's neighbor -> unit
val fmt_print_neighbor: Format.formatter -> 's neighbor -> unit

(** {1 Access to Topological Information}

   When  writing  algorithms, one  can  also  have access  to  topological
   information (i.e., information that are relative to the graph) *)
  
val card : unit -> int
val min_degree : unit -> int
val mean_degree : unit -> float
val max_degree : unit -> int
val is_directed : unit -> bool
val is_cyclic : unit -> bool
val is_connected : unit -> bool
val is_tree : unit -> bool
val height : unit -> (string (* the node id *) -> int) option
val links_number : unit -> int
val diameter : unit -> int

(** It is possible to set some global parameters in the dot file
   using graph attributes. This function allows one the get their
   values. *)
val get_graph_attribute : string -> string


(** {1 Code Registration}

The [register: 's to_register -> unit] function must be called once in
   the user code.  *)

type 's algo_to_register = {
  algo_id   : string;
  init_state: 's state_init_fun;
  enab      : 's enable_fun;
  step      : 's step_fun;
}
type 's to_register = {
    algo : 's algo_to_register list;
    state_to_string:  's  -> string;
    state_of_string: (string -> 's) option;
    copy_state: 's -> 's;
    actions   : action list (** Mandatory in custom daemon mode, or to use oracles *)
  }
(** -  For the [state_to_string] field,  the idea is to  print the raw
   values contained in ['s].  If a  value is omitted, one won't see it
   in the simulation outputs.  If  one prepend a value with "some_id",
   some_id will we  used in the simulation outputs.   Otherwise, an id
   will be invented by sasa.

  - For  the [state_of_string]  field, if  some function  is provided,
   sasa should be able to parse state init values in the dot.
*)

(** To be called once *)
val register : 's to_register -> unit
  
(** {1 Automatic Generation of Registration Code}

   Once enable and step functions are defined, they need to be
   registered by calling the register function.

    An alternative to writing this registration code is to let sasa 
    generate it with the "--gen-register" (or "-reg") option.
    In this case, one needs to follow some naming conventions (w.r.t file 
    and function names):

    (1) The internal state (i.e. the 's shared by all algos) should be 
    defined in a file named "state.ml" that defines the following 5 entities:
     {[
       type t = define_me
       let (to_string: t -> string) = fun _ -> "define_me"
       let (of_string: string -> t option) = fun _ -> None
       let (copy : t -> t) = fun x -> x
       let actions = ["action1";"action2"]; 
     ]}

    If one does not want/need to provide the of_string state parser, returning
    None is fine. This is mandatory only if one wants to define initial values
    in the dot file. 

    Defining a copy that is not the identity is necessary if the state is not 
    functional (e.g., if it contains an array or an Hashtbl).

    In the file "state.ml" does not exist in the current directory, a skeleton 
    is generated.

    (2) All the algos mentioned in the dot file should define the 
    following functions:
     {[
       let (init_state: int -> string -> State.t) = finishme 
       let (enable_f: State.t neighbor list -> State.t -> action list) = finishme
       let (step_f : State.t neighbor list -> State.t -> action -> State.t ) = finishme
     ]}
*)
