(* Time-stamp: <modified the 12/03/2020 (at 16:55) by Erwan Jahier> *)

open Register

let (update_env_with_init : 'v Env.t -> 'v Process.t list -> 'v Env.t) =
  fun e pl ->
    let (aux: 'v Env.t -> 'v Process.t -> 'v Env.t) =
      fun e p ->
        Env.set e p.pid p.init
    in
    List.fold_left aux e pl

(** Returns the channel number that let [p_neighbor] access to the
   content of [p], if [p] is a neighbor of [p_neighbor].  Returns -1 if
   [p] is not a neigbhbor of [p_neigbor], which can happen in directed
   graphs. *)
let (reply: Topology.t -> string -> string -> int) =
  fun g p p_neighbor  ->
    let rec f i = function
      | [] -> (-1) (* may happen in directed graphs *) 
      | (_w,x)::t -> if x=p then i else f (i+1) t
    in
    f 0 (g.succ p_neighbor)

let (get_neighors: Topology.t -> Topology.node_id -> 'v -> 'v Register.neighbor list) =
  fun g source_id init ->
    let idl = g.succ source_id in
    List.map
      (fun (w, neighbor_id) ->
         let node = g.of_id neighbor_id in
         {
           state = init;
           pid = node.id;
           spid = source_id;
           reply = (fun () -> reply g source_id neighbor_id);
           weight = (fun () -> w)
          }
      )
      idl

let (dump_process: 'v Process.t * 'v Register.neighbor list -> unit) =
  fun (p,nl) ->
    let pvars = String.concat "," (SasaState.to_var_names p.pid p.init) in
    let neighbors = List.map StringOf.algo_neighbor nl in
    Printf.eprintf "process %s\n\tvars:%s\n\tneighors: \n\t\t%s\n%!" p.pid  pvars
      (String.concat "\n\t\t" neighbors)

open Process

let (update_env: 'v Env.t -> 'v Process.t * 'v -> 'v Env.t) =
  fun e (p, st) ->
    Env.set e p.pid st    

open SasArg
    
let (update_neighbor_env: 'v Env.t -> 'v Register.neighbor -> 'v Register.neighbor) =
  fun e n ->
    { n with state = Env.get_copy e n.Register.pid }

type 'v layout = ('v Process.t * 'v Register.neighbor list) list

type 'v enable_processes =
  ('v Process.t * 'v Register.neighbor list * Register.action) list list * bool list list

let (get_enable_processes: 'v layout -> 'v Env.t -> 'v enable_processes) =
  fun pl_n e ->
    let all = List.fold_left
        (fun acc (p,nl) ->
           let nl = List.map (update_neighbor_env e) nl in
           let lenv = Env.get_copy e p.pid in
           let al = p.enable nl lenv in
           let al = List.map (fun a -> p,nl,a) al in
           al::acc)
        [] pl_n
    in
    assert (List.length pl_n = List.length all);
    let all = List.rev all in 
    let enab_ll =
      List.map2
        (fun (p,_) al ->
           let al = List.map (fun (_,_,a) -> a) al in
           List.map (fun a_static -> List.mem a_static al) p.actions)
        pl_n
        all
    in
    all, enab_ll

let (do_step : ('v Process.t * 'v Register.neighbor list * action) list -> 'v Env.t
     -> 'v Env.t) =
  fun pnal e -> 
    let lenv_list =
      List.map (fun (p,nl,a) ->
          let nl = List.map (update_neighbor_env e) nl in
          let lenv = Env.get_copy e p.pid in
          p, p.step nl lenv a)
        pnal
    in
    (* 4: update the env *)
    let ne = List.fold_left update_env e lenv_list in
    ne
      
type 'v t = SasArg.t * 'v layout * 'v Env.t


let (get_inputs_rif_decl: SasArg.t -> 'v Process.t list -> (string * string) list) =
  fun args pl -> 
  if args.daemon <> Custom then
    if args.dummy_input then ["_dummy","bool"] else []
  else 
    let f acc p =
      List.fold_left
        (fun acc a -> (p.pid ^(if a="" then "" else "_")^(StringOf.action a) ,"bool")::acc)
        acc
        (List.rev p.actions)
    in
    List.fold_left f [] (List.rev pl)

let (get_outputs_rif_decl: SasArg.t -> 'v Process.t list -> (string * string) list) =
  fun args pl ->
  (* This fonction  may be called on huge lists:  thus it must remains
     tail-recursive and linear! *)
  let pl = List.rev pl in
  let vars = [] in
  (* Adding action vars *)
  let vars =
    if args.daemon = Custom then vars else
      List.fold_left 
        (fun acc p ->
          List.fold_left
            (fun acc a ->
               ((Printf.sprintf "%s_%s" p.pid (StringOf.action a)),"bool")::acc)
            acc
            p.actions
        )
        vars
        pl    
  in
  (* Adding enable action vars *)
  let vars = 
    List.fold_left 
      (fun acc p ->
        List.fold_left
          (fun acc a ->
            ((Printf.sprintf "Enab_%s_%s" p.pid (StringOf.action a)),"bool")::acc)
          acc
          p.actions
      )
      vars
      pl
  in
  (* Adding algo vars *)
  let vars =
    List.fold_left
      (fun acc p ->
        let l = List.rev (SasaState.to_rif_decl p.pid p.init) in
        List.fold_left (fun acc (n,t) -> (n, Data.type_to_string t)::acc) acc l
      )
      vars
      pl
  in
  vars
    
let (env_rif_decl: SasArg.t -> 'v Process.t list -> string) =
  fun args pl ->
    let ssl = get_outputs_rif_decl args pl in
    String.concat " "
      (List.map (fun (base, tstr) -> Printf.sprintf "\"%s\":%s" base tstr) ssl)


let (make : bool -> string array -> 'v t) =
  fun dynlink argv ->
  let args = 
    try SasArg.parse argv;
    with
      Failure(e) ->
      output_string stdout e;
      flush stdout ;
      exit 2
    | e ->
      output_string stdout (Printexc.to_string e);
      flush stdout;
      exit 2
  in
  try
    let dynlink = if args.output_algos then false else dynlink in
    let dot_file = args.topo in
    let g = Topology.read dot_file in
    let nl = g.nodes in
    if args.output_algos then (
      let fl = List.map (fun n -> n.Topology.file) nl in
      let fl = List.sort_uniq compare fl in
      Printf.printf "%s\n%!" (String.concat " " fl);
      exit 0
    );
    let cmxs = (Filename.chop_extension dot_file) ^ ".cmxs" in
    if args.gen_register then (
      let base = Filename.chop_extension dot_file in
      let base = Str.global_replace (Str.regexp "\\.") "" base in
      let ml_register_file = base ^ ".ml" in
      let ml_state_file = "state.ml" in
      let algo_files = List.map (fun n -> n.Topology.file) nl in
      let algo_files = List.sort_uniq compare algo_files in
      let ml_inputs = String.concat " " algo_files in
      GenRegister.f algo_files (ml_state_file, ml_register_file);
      Printf.eprintf " [sasa] Hint: you may wish to generate %s out of %s with:\n"
        cmxs ml_register_file;
      Printf.eprintf " [sasa] ocamlfind ocamlopt -package algo -shared %s %s %s -o %s\n%!"
        ml_state_file ml_inputs ml_register_file cmxs;
      exit 0
    );

    let nidl = List.map (fun n -> n.Topology.id) nl in
    let nstr = String.concat "," nidl in
    Register.set_card (fun () -> List.length nl);
    Register.set_degrees (fun () -> Topology.get_degree g);
    Register.set_mean_deg (fun () -> Topology.get_mean_degree g);
    Register.set_is_connected_cyclic (fun () -> Topology.is_connected_and_cyclic g);
    Register.set_height (Topology.get_height g);
    Register.set_links_number (fun () -> Topology.get_nb_link g);
    Register.set_diameter (fun () -> Diameter.get g);
    Register.set_is_directed (fun () -> g.directed);

    Register.verbose_level := args.verbose;

    if !Register.verbose_level > 1 then Printf.eprintf "==> nodes: %s\n" nstr;

    if dynlink then (
      (* Dynamically link the cmxs file (not possible from rdbg) *)
      let cmxs = Dynlink.adapt_filename cmxs in
      if !Register.verbose_level > 0 then Printf.eprintf " [sasa] Loading %s...\n" cmxs;
      Dynlink.loadfile_private cmxs;
    ) else ();

    let initl = List.map (fun n ->
        if n.Topology.file = "" then (
        Printf.eprintf " [sasa] Empty algo attribute in %s.\n%!" dot_file;
        exit 1
        );
        let algo_id = Filename.chop_suffix n.Topology.file ".ml" in
        let value_of_string_opt = Register.get_value_of_string () in
        if value_of_string_opt = None || n.Topology.init = "" then
          Register.get_init_state algo_id (List.length (g.succ n.id)) n.id
        else
          match value_of_string_opt with
          | None -> assert false (* sno *)
          | Some f -> f n.Topology.init 
      )
        nl
    in

    if !Register.verbose_level > 0 then Printf.eprintf "==> get_neighors\n";
    let algo_neighors = List.map2 (get_neighors g) nidl initl in
    let pl = List.map2 (Process.make (args.daemon=Custom)) nl initl in
    let e = Env.init () in
    let e = update_env_with_init e pl in
    let pl_n = List.combine pl algo_neighors in
    if !Register.verbose_level > 1 then List.iter dump_process pl_n;
    if args.gen_lutin then (
      let fn = (Filename.remove_extension args.topo) ^ ".lut" in
      if Sys.file_exists fn then (
        Printf.eprintf " [sasa] %s already exists: rename it to proceed.\n%!" fn;
        exit 1
      ) else 
        let oc = open_out fn in
        Printf.fprintf oc "%s%!" (GenLutin.f pl); 
        close_out oc;
        Printf.eprintf " [sasa] %s has been generated.\n%!" fn; 
        exit 0);
    if args.gen_oracle then (
      let fn = (Filename.remove_extension args.topo) ^ "_oracle.lus" in
      if Sys.file_exists fn then (
        Printf.eprintf " [sasa] %s already exists: rename it to proceed.\n%!" fn; exit 1
      ) else 
        let oc = open_out fn in
        Printf.fprintf oc "%s%!" (GenOracle.f g pl);
        close_out oc;
        Printf.eprintf " [sasa] %s has been generated.\n%!" fn; 
        exit 0);
    let seed = seed_get args in
    if args.no_data_file then () else (
      let oc = if args.rif then stderr else stdout in
      if !Register.verbose_level > 0 then Printf.eprintf "==> open rif file...\n%!";
      Printf.fprintf oc "%s" (Mypervasives.entete "#" SasaVersion.str SasaVersion.sha);
      Printf.fprintf oc "#seed %i\n" seed;
      let inputs_decl = get_inputs_rif_decl args pl in 
      Printf.printf "#inputs ";
      if !Register.verbose_level > 0 then
        Printf.eprintf "==> get input var names...\n%!";
      List.iter (fun (vn,vt) -> Printf.printf "\"%s\":%s " vn vt) inputs_decl; 
      Printf.printf "\n%!";
      if !Register.verbose_level > 0 then
        Printf.eprintf "==> get output var names...\n%!";
      Printf.printf "#outputs %s\n" (env_rif_decl args pl); 
      Printf.printf "\n%!";
      flush_all()
    );
    if args.ifi then (
      if !Register.verbose_level > 0 then Printf.eprintf "==> read bool...\n%!";
      List.iter
        (fun p -> List.iter
            (fun a -> ignore (RifRead.bool (args.verbose>1) p.pid (StringOf.action a)))
            p.actions)
        pl;
      Printf.eprintf "Ignoring the first vectors of sasa inputs\n%!";
    );
    if !Register.verbose_level > 0 then Printf.eprintf "==> Main.make done !\n%!";
    args, pl_n, e
  with
  | Dynlink.Error e ->
    Printf.eprintf " [sasa] Error when dynlinking (Sasacore.make): %s\n%!"
      (Dynlink.error_message e);
    flush_all();
    exit 2
  | e ->
    Printf.eprintf " [sasa] Error (Sasacore.make): %s\n%!" (Printexc.to_string e);
    flush_all();
    exit 2
