(* Time-stamp: <modified the 06/03/2020 (at 10:20) by Erwan Jahier> *)

type 's neighbor = {
  state:  's ;
  pid: string; 
  spid: string; 
  reply: unit -> int; 
  weight: unit -> int; 
}

type algo_id = string
type action  = string 
type 's enable_fun = 's neighbor list -> 's -> action list
type 's step_fun   = 's neighbor list -> 's -> action -> 's

type 's internal_tables = {
  init_state: (string, Obj.t) Hashtbl.t;
  enable   : (string, Obj.t) Hashtbl.t;
  step     : (string, Obj.t) Hashtbl.t;
  value_to_string  : (string, Obj.t) Hashtbl.t;
  value_of_string  : (string, Obj.t) Hashtbl.t;
  copy_value  : (string, Obj.t) Hashtbl.t;
  graph_attributes  : (string, string) Hashtbl.t;
  mutable actions:action list;
  mutable card : int ;
  mutable min_deg      : int;
  mutable mean_deg     : float; 
  mutable max_deg      : int;
  mutable is_cyclic    : bool option;
  mutable is_connected : bool option;
  mutable is_tree      : bool option;
  mutable is_directed  : bool option;
  mutable height       : (string -> int); 
  mutable links_number : int; 
  mutable diameter     : int;
  }

type properties_functions = {
  mutable card                : unit -> int;
  mutable min_max             : unit -> int*int;
  mutable mean_deg            : unit -> float;
  mutable is_connected_cyclic : unit -> bool*bool;
  mutable is_directed         : unit -> bool;
  mutable links_number        : unit -> int;
  mutable diameter            : unit -> int
}

type node_id = string (* cf topology.mli *)

let (tbls:'s internal_tables) = {
  init_state = Hashtbl.create 1;
  enable    = Hashtbl.create 1;
  step      = Hashtbl.create 1;
  value_to_string   = Hashtbl.create 1;
  value_of_string   = Hashtbl.create 1;
  copy_value   = Hashtbl.create 1;
  graph_attributes = Hashtbl.create 1;
  actions   = [];
  card         = (-1);
  min_deg      = (-1);
  mean_deg     = (-1.);
  max_deg      = (-1);
  is_cyclic    = None;
  is_connected = None;
  is_tree      = None;
  is_directed  = None;
  height       = (fun _ -> -1);
  links_number = (-1);
  diameter     = (-1)
}

let (prop_funs:properties_functions) = {
  card                = (fun () -> -1);
  min_max             = (fun () -> (-1,-1));
  mean_deg            = (fun () -> -1.);
  is_connected_cyclic = (fun () -> (false,false));
  is_directed         = (fun () -> false);
  links_number        = (fun () -> -1);
  diameter            = (fun () -> -1)
}

let verbose_level = ref 0
exception Unregistred of string * string

let print_table lbl tbl =
  let keys = Hashtbl.fold (fun k _ acc -> Printf.sprintf "%s,%s" k acc)  tbl "" in
  if !verbose_level > 0 then Printf.eprintf "Defined keys for %s: %s\n%!" lbl keys
    

let (reg_init_state : algo_id -> (int -> string -> 's) -> unit) =
  fun algo_id x ->
    if !verbose_level > 0 then Printf.eprintf "Registering %s init_vars\n%!" algo_id;
    Hashtbl.replace tbls.init_state algo_id (Obj.repr x)

    
let (get_init_state : algo_id -> int -> string -> 's) =
  fun algo_id ->
    try Obj.obj (Hashtbl.find tbls.init_state algo_id)
    with Not_found ->
      print_table "init_state" tbls.init_state;
      raise (Unregistred ("init_state", algo_id))

      
let (reg_enable : algo_id -> 's enable_fun -> unit) = fun algo_id x -> 
  if !verbose_level > 0 then Printf.eprintf "Registering %s enable\n%!" algo_id;
  Hashtbl.replace tbls.enable algo_id (Obj.repr x)
let (get_enable : algo_id -> 's enable_fun) = fun algo_id -> 
  try Obj.obj (Hashtbl.find tbls.enable algo_id)
  with Not_found ->
    print_table "enable" tbls.enable;
    raise (Unregistred ("enable", algo_id))

let (reg_step : algo_id -> 's step_fun -> unit) = fun algo_id x -> 
  if !verbose_level > 0 then Printf.eprintf "Registering %s step\n%!" algo_id;
  Hashtbl.replace tbls.step algo_id (Obj.repr x)
    
let (get_step : algo_id -> 's step_fun) = fun algo_id -> 
  try Obj.obj (Hashtbl.find tbls.step algo_id)
  with Not_found ->
    print_table "step" tbls.step;
    raise (Unregistred ("step", algo_id))

let (reg_actions : action list -> unit) =
  fun x -> 
  if !verbose_level > 0 then Printf.eprintf "Registering actions\n%!";
   tbls.actions <- x
let (get_actions : unit -> action list) = fun () -> 
  tbls.actions 

let (reg_value_to_string : ('s -> string) -> unit) =
  fun f -> 
  if !verbose_level > 0 then Printf.eprintf "Registering value_to_string\n%!";
  Hashtbl.replace tbls.value_to_string "_global" (Obj.repr f)
let (get_value_to_string : unit -> 's -> string) = fun () -> 
  try Obj.obj (Hashtbl.find tbls.value_to_string "_global")
  with Not_found ->
    print_table "value_to_string" tbls.value_to_string;
    raise (Unregistred ("value_to_string", "_global"))

let (reg_value_of_string : (string -> 's) -> unit) =
  fun f -> 
  if !verbose_level > 0 then Printf.eprintf "Registering value_of_string\n%!";
  Hashtbl.replace tbls.value_of_string "_global" (Obj.repr f)
let (get_value_of_string : unit -> (string -> 's) option) = fun () -> 
  try Some (Obj.obj (Hashtbl.find tbls.value_of_string "_global"))
  with Not_found -> None


let (reg_copy_value : ('s -> 's) -> unit) =
  fun f -> 
  if !verbose_level > 0 then Printf.eprintf "Registering copy_value\n%!";
  Hashtbl.replace tbls.copy_value "_global" (Obj.repr f)
    
let (get_copy_value : unit ->  ('s -> 's)) = fun () -> 
  try Obj.obj (Hashtbl.find tbls.copy_value "_global")
  with Not_found ->
    print_table "copy_value" tbls.copy_value;
    raise (Unregistred ("copy_value", "_global"))


let set_min_max : (unit -> unit) = 
  fun () -> 
    let (x,y) = prop_funs.min_max () in
    tbls.min_deg <- x;
    tbls.max_deg <- y

let set_connec_cycl : (unit -> unit) = 
  fun () ->
    let (x,y) = prop_funs.is_connected_cyclic () in
    tbls.is_connected <- Some x;
    tbls.is_cyclic <- Some y

let (card : unit -> int) = fun () ->
  match tbls.card with 
  | -1 ->
     let c = prop_funs.card () in
     tbls.card <- c;
     c
  | c -> c

let (is_directed : unit -> bool) = fun () ->
  match tbls.is_directed with 
  | None ->
     let c = prop_funs.is_directed () in
     tbls.is_directed <- Some c;
     c
  | Some c -> c

let (min_degree : unit -> int) = 
  fun () -> match tbls.min_deg with 
  | -1 -> (set_min_max (); tbls.min_deg)
  | m -> m

let (mean_degree : unit -> float) = fun () ->
  match tbls.mean_deg with 
  | -1. ->
     let m = prop_funs.mean_deg () in
     tbls.mean_deg <- m;
     m
  | m -> m

let (max_degree : unit -> int) = 
  fun () -> match tbls.max_deg with 
  | -1 -> (set_min_max (); tbls.max_deg)
  | m -> m

let (is_cyclic : unit -> bool) = fun () ->
  match tbls.is_cyclic with 
  | None ->
     set_connec_cycl (); 
     (match tbls.is_cyclic with 
      | Some b -> b 
      | _ -> assert false)
  | Some b -> b

let (is_connected : unit -> bool) = fun () ->
  match tbls.is_connected with 
  | None -> (
    set_connec_cycl (); 
    match tbls.is_connected with 
    | Some b -> b 
    | _ -> assert false)
  | Some b -> b

let (is_tree : unit -> bool) = fun () ->
  match tbls.is_tree with 
  | None ->
     let b = (not (is_cyclic ()) && (is_connected ())) in
     tbls.is_tree <- Some b;
     b
  | Some b -> b

(* Caution : this option is not the same as the option in the type tbls.height. 
 * If height () = None, then the graph doesn't have a height (because it isn't a tree)*)
let height : (unit -> (string -> int) option) = 
  fun () -> 
    if is_tree () then
      Some tbls.height
    else None


let (links_number : unit -> int) = 
  fun () ->
  match tbls.links_number with 
  | -1 ->
     let n = prop_funs.links_number () in
     tbls.links_number <- n;
     n
  | n -> n

let (diameter : unit -> int) = 
  fun () -> 
  match tbls.diameter with
  | -1 ->
     let d = (prop_funs.diameter ()) in
     tbls.diameter <- d;
     d
  | d -> d
  


let set_card : ((unit -> int) -> unit) =
  fun f -> prop_funs.card <- f

let set_is_directed : ((unit -> bool) -> unit) =
  fun f -> prop_funs.is_directed <- f

let set_degrees : ((unit -> int*int) -> unit) = 
  fun f -> prop_funs.min_max <- f

let set_mean_deg : ((unit -> float) -> unit) = 
  fun f -> prop_funs.mean_deg <- f

let set_is_connected_cyclic : ((unit -> bool*bool) -> unit) = 
  fun f -> prop_funs.is_connected_cyclic <- f

let set_height : ((node_id -> int) -> unit) = 
  fun f -> tbls.height <- f

let set_links_number : ((unit -> int) -> unit) = 
  fun f -> prop_funs.links_number <- f

let set_diameter : ((unit -> int) -> unit) = 
  fun f -> prop_funs.diameter <- f


let (to_string : 's -> string) =
  fun v ->
    (get_value_to_string ()) v

let (get_graph_attribute : string -> string) =
  fun str ->
    try Hashtbl.find tbls.graph_attributes str 
    with Not_found ->
      failwith (Printf.sprintf "The graph attribute %s does not seem to exist" str)

let (set_graph_attribute : string -> string -> unit) =
  Hashtbl.replace tbls.graph_attributes 

let (graph_attribute_list: unit -> (string * string)  list) =
  fun () ->
    Hashtbl.fold (fun n v acc -> (n,v)::acc) tbls.graph_attributes []
