(* This file defines sasa specific commands. 

It is meant  to be loaded after the (rdbg-generated) "rdbg-cmds.ml" file.
Look at the sasa/test/*/my-rdbg-commands.ml files for examples
*)

#require "sasa";;
       
(**********************************************************************)
(* Dealing with rounds *)
let roundnb = ref 1
let roundtbl = Hashtbl.create 1;;


(**********************************************************************)

(* BEWARE: The paths below are specific to sasa and to the sasa test directory *)

#mod_use "../../lib/algo/algo.ml";; 
#mod_use "../../lib/sasacore/register.ml";;
#mod_use "../../lib/sasacore/topology.ml";;
#use "../rdbg-utils/dot.ml";;

let p = Topology.read dotfile;;


(* Shortcuts for dot viewers defined in rdbg-utils/dot.ml *)
let d_par () = dot true !roundnb p dotfile !e;;
let dot_par () = dot true !roundnb p dotfile !e;;
let ne_par () = neato true !roundnb p dotfile !e;;
let tw_par () = twopi true !roundnb p dotfile !e;;
let ci_par () = circo true !roundnb p dotfile !e;;
let fd_par () = fdp true !roundnb p dotfile !e;;
let sf_par () = sfdp true !roundnb p dotfile !e;;
let pa_par () = patchwork true !roundnb p dotfile !e;;
let os_par () = osage true !roundnb p dotfile !e;;

let dot () = dot false !roundnb p dotfile !e;;
let ne () = neato false !roundnb p dotfile !e;;
let tw () = twopi false !roundnb p dotfile !e;;
let ci () = circo false !roundnb p dotfile !e;;
let fd () = fdp false !roundnb p dotfile !e;;
let sf () = sfdp false !roundnb p dotfile !e;;
let pa () = patchwork false !roundnb p dotfile !e;;
let os () = osage false !roundnb p dotfile !e;;


(* To change the default dot/graph viewer: 
   dot_view := ci;;
*)
let dot_view : (unit -> unit) ref =
  ref ( (* Try to chose the most sensible default viewer *)
      if String.length dotfile > 4 && String.sub dotfile 0 4 = "ring" then ci else
      if Algo.is_directed () then dot else ne)

let d () = !dot_view ()
let sd () = s();!dot_view();;
let nd () = n();!dot_view();;
let bd () = b();!dot_view();;


(**********************************************************************)
(** Computing rounds *)

(* a process can be removed from the mask if one action of p is triggered
  or if no action of p is enabled *)
let get_removable pl =
  let pl = List.filter
      (fun p ->
         (List.exists (fun (_,_,acti) -> acti) p.actions) ||
         (List.for_all (fun (_,enab,_) -> (not enab)) p.actions)
      )
      pl
  in
  List.map (fun p -> p.name) pl

let verbose = ref false

let last_round = ref 0
let mask = ref [] (* nodes we look the activation of *)


(* called at each event via the time-travel hook *)
let (round : RdbgEvent.t -> bool) =
  fun e ->
  try
    let croundnb, round = Hashtbl.find roundtbl e.nb in
    (*     Printf.printf "round tabulated at e.nb %d: croundnb=%d round = %b\n%!" *)
    (*       e.nb croundnb round; *)
    roundnb := croundnb;
    round
  with Not_found ->
    let round =
      e.kind = Ltop &&
      let (pl : process list) = get_processes e in
      let rm_me = get_removable pl in
      if !verbose then (
        Printf.printf "Mask : %s\n" (String.concat "," !mask); 
        Printf.printf "To remove from mask: %s\n" (String.concat "," rm_me); 
        flush stdout;
      );
      mask := List.filter (fun pid -> not (List.mem pid rm_me)) !mask;
      let res = !mask = [] ||
                (* when round is called twice, it should have the same
                     result *) 
                !last_round = e.nb 
      in
      if !mask = [] then (
        last_round := e.nb;
        mask := (
          let p_with_enable_action =
            List.filter
              (fun p -> List.exists
                  (fun (_,enab,acti) -> enab && not(acti)) p.actions)
              pl
          in
          let pidl = List.map (fun p -> p.name) p_with_enable_action in
          let pidl = List.rev pidl in
          if !verbose then (
            Printf.printf "Next mask : %s\n" (String.concat "," pidl);
            flush stdout
          );
          pidl
        )
      );
      res
    in
    if round && e.nb > 1 then incr roundnb;
    Hashtbl.add roundtbl e.nb (!roundnb, round);
    (* Printf.printf "round computed at e.nb %d: croundnb=%d round = %b\n%!"
       e.nb !roundnb round; *)
    round

let update_round_nb e =
  match Hashtbl.find_opt roundtbl e.nb with
  | None -> roundnb := 1
  | Some (n,_) -> roundnb := n

(* go to next and previous rounds  *)
let next_round e = next_cond e round;;
let nr () = e:=next_round !e; store !e.nb; !dot_view ();;
let pr () =
  e:=goto_last_ckpt !e.nb;
  update_round_nb !e;
  !dot_view ();
  store !e.nb

(* I need to overrides those *)

let u () = undo (); ignore (round !e);;
let r () = r (); roundnb := 1; Hashtbl.clear roundtbl;ignore (round !e);;

(**********************************************************************)
(* print_event tuning *)

(* split  the vars into returns  (the enab vars, the  activated vars,
   the other vars) nb: in the "Enab" prefix is removed from enab vars
   names; ie we leave only the pid and the action name *)
type s = (string * string * Data.v) 
let split_data (l:Data.subst list) : s list * s list * s list = 
    let l = List.map (fun (x,v) -> Str.split (Str.regexp "_") x, v) l in
    let rec sortv (enab, other) (x,v) =
      match x with 
      | "Enab"::pid::tail -> (pid, String.concat "_" tail,v)::enab, other
      | pid::tail -> enab, (pid,(String.concat "_" tail),v)::other
      | [] -> assert false
    in
    let enab, other = List.fold_left sortv ([],[]) l in
    let acti_names = List.map (fun (pid, n, v) -> pid, n) enab in  
    let act, vars =
      List.partition (fun (pid,n, _) -> List.mem (pid,n) acti_names) other
    in
    enab, act, vars


let only_true l = List.filter (fun (_,_, v)  -> v = B true) l
(* Only print the active process values *)
let str_of_sasa_event short e =
    let enab, act, vars = split_data e.data in
    (*     let enab = only_true enab in *)
    let act = only_true act in
    let act_pid = List.map (fun (pid,_,_) -> pid) act in
    let vars = List.filter (fun (pid, _,_) -> List.mem pid act_pid) vars in
    let _to_string (pid, n, _) = Printf.sprintf "%s_%s" pid n in 
    let to_string_var (pid, n, v) =
      Printf.sprintf "%s_%s=%s" pid n (Data.val_to_string string_of_float v)
    in
    let vars = List.rev vars in
    update_round_nb e;
    if short then
      Printf.sprintf "[round %i, step %i%s] %s\n" !roundnb e.step
        (if e.step <> e.nb then (":" ^ (string_of_int e.nb)) else "")
        (String.concat " " (List.map to_string_var vars))
    else
      Printf.sprintf "Round %i - Step %i%s \nActive node states: %s\n" !roundnb e.step
        (if e.step <> e.nb then ("\nEvent nb " ^ (string_of_int e.nb)) else "")
        (String.concat " " (List.map to_string_var vars))
      
      
let print_sasa_event short e =
  if e.kind <> Ltop then print_event e else
    (
      print_string (str_of_sasa_event short e);
      flush stdout
    )

let _ =
  del_hook "print_event";
  add_hook "print_event" (print_sasa_event true)

(* shortcuts to the default and sasa event printers *)
let pe () = print_event !e;;
let spe () = print_sasa_event true !e;;
let spel () = print_sasa_event false !e;;

(**********************************************************************)
(* Goto to the next Lustre oracle violation *)
let goto_next_false_oracle e =
  next_cond e (fun e -> e.kind = Exit && e.lang = "lustre" &&
                        List.mem ("ok", Bool) e.outputs  &&
                        not (vb "ok" e))

let viol () =
  if args.oracles <> [] then (
    e:=goto_next_false_oracle !e; !dot_view ()
  ) else (
    Printf.printf "No oracle is set.\n%!"
  )
;;

(**********************************************************************)
(* Move forward until silence *)

let is_silent e =
  let enab, _act, _vars = split_data e.data in
  List.for_all (fun (_,_,v) -> v = Data.B false) enab

let goto_silence e = next_cond e is_silent 
let silence () = e:=goto_silence !e; !dot_view ();;

let _ = add_doc_entry
    "silence" "unit -> unit" "Move forward until is_silent returns true"
    "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry
    "is_silent" "RdbgEvent.t -> bool"
    "is the event correspond to a silent configuration? (i.e., no enable node)"
    "sasa" "sasa-rdbg-cmds.ml";;


let legitimate = silence;;
let _ = add_doc_entry
    "legitimate" "unit -> unit"
    "  Move forward until a legitimate configuration is reached (uses 'silence' by default)"
    "sasa" "sasa-rdbg-cmds.ml";;
  

(**********************************************************************)
(* Perform the checkpointing at rounds! *)
let _ = check_ref := round;;

(**********************************************************************)
let _ = 
  add_doc_entry "d" "unit -> unit" "update the current network with dot" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "ne" "unit -> unit" "update the current network with neato" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "tw" "unit -> unit" "update the current network with twopi" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "ci" "unit -> unit" "update the current network with circo" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "fd" "unit -> unit" "update the current network with fdp" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "sf" "unit -> unit" "update the current network with sfdp" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "pa" "unit -> unit" "update the current network with patchwork" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "os" "unit -> unit" "update the current network with osage" "sasa" "sasa-rdbg-cmds.ml";
  
  add_doc_entry "sd" "unit -> unit"
    "go to the next step and update the network with one of the GraphViz tools" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "nd" "unit -> unit" "go to the next event and update the network" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "bd" "unit -> unit" "go to the previous event and update the network" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "nr" "unit -> unit" "go to the next round and update the network" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "pr" "unit -> unit" "go to the previous round and update the network" "sasa" "sasa-rdbg-cmds.ml"
  ;;

let l () =
  l();
  print_string ("
- sasa commands
 + Display network with GraphViz tools
   d: "^(RdbgMain.doc_msg "d")^"
   ne: "^(RdbgMain.doc_msg "ne")^"
   tw: "^(RdbgMain.doc_msg "tw")^" 
   ci: "^(RdbgMain.doc_msg "ci")^"
   fd: "^(RdbgMain.doc_msg "fd")^"
   sf: "^(RdbgMain.doc_msg "sf")^"
   pa: "^(RdbgMain.doc_msg "pa")^"
   os: "^(RdbgMain.doc_msg "os")^"
   
   nb: for algorithms that have a field named 'par', you can try 
   of the following (which only draw the parent arcs) 
 
   d_par: "^(RdbgMain.doc_msg "d")^" (parent arcs only)
   ne_par: "^(RdbgMain.doc_msg "ne")^" (parent arcs only)
   tw_par: "^(RdbgMain.doc_msg "tw")^" (parent arcs only)
   ci_par: "^(RdbgMain.doc_msg "ci")^" (parent arcs only)
   fd_par: "^(RdbgMain.doc_msg "fd")^" (parent arcs only)
   sf_par: "^(RdbgMain.doc_msg "sf")^" (parent arcs only)
   pa_par: "^(RdbgMain.doc_msg "pa")^" (parent arcs only)
   os_par: "^(RdbgMain.doc_msg "os")^" (parent arcs only)

 + Moving commands [*]
   sd: "^(RdbgMain.doc_msg "sd")^"
   nd: "^(RdbgMain.doc_msg "nd")^"
   bd: "^(RdbgMain.doc_msg "bd")^"
   nr: "^(RdbgMain.doc_msg "nr")^"
   pr: "^(RdbgMain.doc_msg "pr")^"
 [*] in order to change the current graph drawing engine, you can use
the dot_view command as follows:
  dot_view := ci
");;


(**********************************************************************)
(* ok, let's start the debugging session! *)

let pdf_viewer =
  if Sys.command "which zathura" = 0 then "zathura" else
  if Sys.command "which xpdf" = 0 then "xpdf" else
  if Sys.command "which acroread" = 0 then "acroread" else
  if Sys.command "which evince" = 0 then "evince" else (
    Printf.printf "Warning: no pdf viewer is found to visualize %s\n%!" dotfile;
    "ls"
  )
  
let graph_view () = 
  !dot_view ();
  let cmd = Printf.sprintf "%s sasa-%s.pdf&" pdf_viewer dotfile in
  Printf.printf "%s\n!" cmd;
  ignore(Sys.command cmd); 
  ignore(round !e)

 ;;
(* let _ = graph_view () *)
let gv = graph_view

let _ = 
  add_doc_entry "gv" "unit -> unit" "launch a pdf viewer of the current network"
    "sasa" "sasa-rdbg-cmds.ml";
