open Sasacore
open Topology
open Ggcore
open List

type node_succ_t = (string, (int * string) list) Hashtbl.t

let nid_list_remove : (node_id list -> node_id -> (int*node_id) list) =
  fun l e ->
  rev (fold_left (fun acc elem -> if(elem <> e) then (1,elem)::acc else acc ) [] l)


let (gen_clique: bool -> int -> Topology.t) =
  fun directed nb ->
  let (node_succ:node_succ_t) = Hashtbl.create nb
  and nodes = create_nodes "p" (0,nb)
  in
  List.iter
    (fun node_id -> Hashtbl.replace node_succ node_id (nid_list_remove nodes node_id))
    nodes;
  let nl = id_to_empty_nodes nodes in
  {
    nodes = nl;
    succ = (fun n -> try Hashtbl.find node_succ n with Not_found -> []);
    of_id = get_of_id nl;
    directed = directed
  }


let (gen_star: bool -> int -> Topology.t) =
  fun directed nb ->
  let (node_succ:node_succ_t) = Hashtbl.create nb and nodes = "root"::(create_nodes "p" (1,nb)) in
  let first = hd nodes in
  List.iter
    (fun node ->
       Hashtbl.replace node_succ node
         (if node = first then nid_list_remove nodes node else [(1,first)])) nodes;
  let nl = id_to_empty_nodes nodes in
  {
    nodes = nl;
    succ = (fun n -> try Hashtbl.find node_succ n with Not_found -> []);
    of_id = get_of_id nl;
    directed = directed
  }

let add_weight (li : node_id list) : (int * node_id) list = map (fun elem -> (1,elem)) li

let neighbours_ring : (node_id list -> (node_id -> (int * node_id) list)) = 
  fun li ->
  let node_succ:node_succ_t = Hashtbl.create (length li) in
  let ((_,last),ret) = fold_left (fun ((first,prev),accu) elem -> 
    if first = "" then 
      ((elem,elem),[[""; ""]]) 
    else 
      (
        match accu with 
        | [x;_]::tl -> ((first,elem),[prev;first]::([x;elem]::tl))
        | _ -> assert false
      )
  ) (("",""),[]) li in let ret = rev ret in
  let ret = (match ret with
  | ["";x]::tail -> [last;x]::tail
  | _ -> assert false) in
  iter2 (fun neighbours elem -> Hashtbl.replace node_succ elem (add_weight neighbours)) ret li ;
  (fun n -> try Hashtbl.find node_succ n with Not_found -> [])

let (gen_ring: bool -> int -> Topology.t) =
  fun directed nb ->
  let nodes = (create_nodes "p" (0,nb)) in
  let nl = id_to_empty_nodes nodes in
  {
    nodes = nl;
    succ = neighbours_ring nodes;
    of_id = get_of_id nl;
    directed = directed
  }


let (gen_grid: bool -> int -> int -> Topology.t) =
  fun directed length width -> 
  Printf.eprintf "Computing a %ix%i grid...\n" length width;
  flush stderr;
  let nb = length*width in
  let nodes = (create_nodes "p" (0,nb)) and table = Hashtbl.create nb in
  for i=0 to length-1 do
    for j=0 to width-1 do
      let n_id = (List.nth nodes (j*length + i)) in
      let bl = if(i=0) then 0 else -1 in
      let br = if(i=(length-1)) then 0 else 1 in
      let bup = if(j=0) then 0 else -1 in
      let bdown = if(j=(width-1)) then 0 else 1 in
      for ip=bl to br do
        for jp=bup to bdown do
          if not ((ip=0 && jp=0) || (ip=jp) || (ip = -jp)) then
            (Hashtbl.replace table n_id
               ((1,(List.nth nodes ((j+jp)*length + i+ip)))::(
                   try Hashtbl.find table n_id with Not_found -> [])); ) else () 
        done;
      done;
    done;
  done;
  let nl = id_to_empty_nodes nodes in
  Printf.eprintf "Computing a %ix%i grid: Done!\n" length width;flush stderr;
  {
    nodes = nl;
    succ = (fun nid -> (try Hashtbl.find table nid with Not_found -> []));
    of_id = get_of_id nl;
    directed = directed
  }

let rec link_hypercube_nodes : (node_id array -> node_succ_t -> unit) =
  fun na n_s -> 
  let len = Array.length na in let mid = len / 2 in
  if len > 1 then 
    let n1 = (Array.sub na 0 mid) and n2 = (Array.sub na mid mid) in
    link_hypercube_nodes n1 n_s; 
    link_hypercube_nodes n2 n_s;
    Array.iter2 (fun node1 node2 -> 
        Hashtbl.replace n_s node1
          (( 1,node2)::(try Hashtbl.find n_s node1 with Not_found -> []));
        Hashtbl.replace n_s node2
          ((1,node1)::(try Hashtbl.find n_s node2 with Not_found -> []))
      ) n1 n2

let neighbours_hyper_cube : (node_id list -> (node_id -> (int * node_id) list)) = 
  fun nl -> 
    let na = Array.of_list nl in 
    let (node_succ:node_succ_t) = Hashtbl.create (Array.length na) in
    link_hypercube_nodes na node_succ;
    (fun n -> try Hashtbl.find node_succ n with Not_found -> [])

let gen_hyper_cube : (bool -> int -> Topology.t) = 
  fun directed dim -> 
    let nb = int_of_float (2. ** (float_of_int dim)) in
    let nodes = (create_nodes "p" (0,nb)) in
    let nl = id_to_empty_nodes nodes in
    {
      nodes = nl;
      succ = neighbours_hyper_cube nodes;
      of_id = get_of_id nl;
      directed = directed
  }
