open Sasacore
open Topology
open Ggcore
open List

type node_succ_t = (node_id, (int * node_id) list) Hashtbl.t
type probability = float (*between 0 and 1*)

let gen_ER : (bool -> int -> probability -> Topology.t) =
  fun directed nb p ->
  let (node_succ:node_succ_t) = Hashtbl.create nb
  and nodes = create_nodes "p" (0,nb)
  in
  let succ n = try Hashtbl.find node_succ n with Not_found -> [] in
  let add_succ n m = Hashtbl.replace node_succ n ((1,m)::(succ n)) in
  iteri (fun i n ->
      iteri (fun j m ->
          if not directed then (
            if i < j && (Random.float 1.) < p then (
              add_succ n m;
              add_succ m n
            );
          ) else (
            if i <> j && (Random.float 1.) < p then
              add_succ n m;
          )
        )
        nodes
    )
    nodes;
  let nl = id_to_empty_nodes nodes in
  {
    nodes = nl;
    succ = succ;
    of_id = get_of_id nl;
    directed = directed
  }


let rec init_m_nodes : (int -> node_succ_t -> node_id list -> node_id list) =
  fun i node_succ ->
  function
  | (node::tail) ->
    if i > 0 then
      (Hashtbl.replace node_succ node [];
            init_m_nodes (i-1) node_succ tail)
    else node::tail
  | _ -> assert false

let (neighbours_BA : node_id list -> int -> node_succ_t ->
     (node_id -> (int * node_id) list)) =
  fun nodes m node_succ ->
  let d_tot = 2 * m in
  let nodes = init_m_nodes m node_succ nodes in
  match nodes with
  | [] -> assert false
  | head::nodes ->
    Hashtbl.replace node_succ head
      (Hashtbl.fold
          (fun n _ succ ->
             Hashtbl.replace node_succ n [(1,head)];
             (1,n)::succ) node_succ []
      );
    (*init terminée. On a un graph connexe pour les m+1 premiers points,
      nl ne contient que les points non ajoutés*)
    ignore (
      fold_left
        (fun deg_tot node ->
           let deg_temp = deg_tot in
           let succ = ref [] in
           let deg_temp = ref deg_temp in
           for _ = 0 to m-1 do (*for each edge to create*)
             let ran = Random.int !deg_temp in
             ignore (
               Hashtbl.fold
                 (fun n_id n_succ r ->
                    if r >= 0 && not (List.mem (1,n_id) !succ) then
                      let r = r - (length n_succ) in 
                      if r < 0 then (
                        succ := (1,n_id)::!succ;
                        Hashtbl.replace node_succ n_id
                          ((1,node)::n_succ);
                        deg_temp := !deg_temp - length n_succ
                      );
                      r
                    else r
                 )
                 node_succ ran);
           done;
           Hashtbl.replace node_succ node !succ;
           (deg_tot + (2 * m))
        )
        d_tot
        nodes
    );
    (fun n -> try Hashtbl.find node_succ n with Not_found -> [])

let gen_BA : (bool -> int -> int -> Topology.t) =
  fun directed nb m ->
  if directed then (
    Printf.eprintf "A Barabasi–Albert graph cannot be directed\n%!";
    exit 2
  );
  let (node_succ:node_succ_t) = Hashtbl.create nb
  and nodes = create_nodes "p" (0,nb) in
    if nb < m + 1 then 
     (Printf.eprintf
        "Error: with -m %d, the node number needs to be at least %d (it is %d).\n%!"
        m (m+1) nb;
      exit 2
    );
    let nl = id_to_empty_nodes nodes in
    {
      nodes = nl;
      succ = neighbours_BA  nodes m node_succ;
      of_id = get_of_id nl;
      directed = directed
    }

let pre_rand_tree : (node_succ_t -> node_id list -> (node_id -> (int * node_id) list)) =
  fun node_succ ->
  function
  | [] -> failwith "Tree Error : You need at least one nodes in your tree"
  | h::t ->
    ignore (List.fold_left (fun acc elem ->
        let no = (List.nth acc (Random.int (List.length acc))) in
        (Hashtbl.replace node_succ no ((1,elem)::(try Hashtbl.find node_succ no with Not_found -> []));
        Hashtbl.replace node_succ elem ((1,no)::(try Hashtbl.find node_succ elem with Not_found -> []))
        );
         (elem::acc)
    ) [h] (t));
    (fun n -> try Hashtbl.find node_succ n with Not_found -> [])

let (rand_tree: bool -> int -> Topology.t) =
  fun directed nb ->
    let (node_succ:node_succ_t) = Hashtbl.create nb and nodes = create_nodes "p" (0,nb) in
    let nl = id_to_empty_nodes nodes in
    {
      nodes = nl;
      succ = (pre_rand_tree node_succ nodes);
      of_id = get_of_id nl;
      directed = directed
    }


type node_udg =  node_id*float*float
type plan_udg = node_udg list

let (make_plan_udg: node_id list -> float -> float -> plan_udg) =
  fun nodes x y ->
  List.map (fun elem -> (elem,(Random.float x),(Random.float y))) (nodes)

let (dist_udg: node_udg -> node_udg -> float) =
  fun n1 n2 ->
  let (_,x1,y1) = n1 and (_,x2,y2) = n2 in
  sqrt (((x1-.x2)**2.) +. ((y1 -. y2)**2.))

let gen_qudg : (bool -> int -> float -> float -> float -> float -> float ->
                (Topology.t * plan_udg)) =
  fun directed nb x y r0 r1 p ->
  let (node_succ:node_succ_t) = Hashtbl.create nb and nodes = create_nodes "p" (0,nb) in
  let pl = (make_plan_udg nodes x y) in
  List.iter (fun n_udg ->
            let (node, _, _) = n_udg in
            List.iter (fun elem ->
                let (n,_,_) = elem and dist = dist_udg n_udg elem in

                if node <> n && (dist <= r0 || (dist <= r1 && Random.float 1. <= p)) 
                (* e.q. if the node is : (within the radius r0) 
                or : (within the radius r1, with a brobability of p) *)
                then (
                  Hashtbl.replace node_succ node 
                    ((1,n)::(try Hashtbl.find node_succ node with Not_found -> []))
                )
            ) pl
  ) pl;
  let nl = id_to_empty_nodes nodes in
  {
    nodes = nl;
    succ =(fun n -> (try Hashtbl.find node_succ n with Not_found -> []));
    of_id = get_of_id nl;
    directed = directed
  },pl

let gen_udg : (bool -> int -> float -> float -> float -> (Topology.t * plan_udg)) =
  fun directed nb x y r -> (gen_qudg directed nb x y r 0. 0.)
